@@ -7,15 +7,20 @@ from . import naive_bayes
 from . import nearest_neighbors
 from . import sklearn
 from . import util
-from .data_ingest import InputOutputData
+from .data import InputOutputData
 from .data_transformation import DataFrameTransformer, RuleBasedDataFrameTransformer
 from .ensemble import AveragingVectorRegressionModel
 from .evaluation.eval_stats import eval_stats_classification, eval_stats_regression
 from .normalisation import NormalisationMode
-from .vector_model import VectorModel, VectorRegressionModel, VectorClassificationModel
+from .tensor_model import TensorToTensorRegressionModel, TensorToScalarRegressionModel, \
+    TensorToTensorClassificationModel, TensorToScalarClassificationModel
+from .vector_model import PredictorModel, VectorModel, VectorRegressionModel, VectorClassificationModel
+
+__version__ = "0.1.0"
 
 # The following submodules are not imported by default to avoid necessarily requiring their dependencies:
 # tensorflow
 # torch
 # lightgbm
 # catboost
+# geoanalytics
