@@ -1,4 +1,6 @@
+from datetime import datetime
 from concurrent.futures import ProcessPoolExecutor
+import uuid
 
 import logging
 import os
@@ -11,7 +13,7 @@ from typing import Dict, Sequence, Any, Callable, Generator, Union, Tuple, List,
 from .evaluation.evaluator import MetricsDictProvider
 from .local_search import SACostValue, SACostValueNumeric, SAOperator, SAState, SimulatedAnnealing, \
     SAProbabilitySchedule, SAProbabilityFunctionLinear
-from .tracking.tracking_base import TrackedExperimentDataProvider, TrackedExperiment
+from .tracking.tracking_base import TrackingMixin
 from .vector_model import VectorModel
 
 log = logging.getLogger(__name__)
@@ -112,18 +114,29 @@ class ParametersMetricsCollection:
     """
     Utility class for holding and persisting evaluation results
     """
-    def __init__(self, csvPath=None, sortColumnName=None, ascending=True):
+    def __init__(self, csvPath=None, sortColumnName=None, ascending=True, incremental=False):
         """
         :param csvPath: path to save the data frame to upon every update
         :param sortColumnName: the column name by which to sort the data frame that is collected; if None, do not sort
         :param ascending: whether to sort in ascending order; has an effect only if sortColumnName is not None
+        :param incremental: whether to add to an existing CSV file instead of overwriting it
         """
         self.sortColumnName = sortColumnName
         self.csvPath = csvPath
         self.ascending = ascending
-        self.df = None
-        self.cols = None
-        self._currentRow = 0
+        if os.path.exists(csvPath) and incremental:
+            self.df = pd.read_csv(csvPath)
+            log.info(f"Found existing CSV file with {len(self.df)} entries; {csvPath} will be extended (incremental mode)")
+            self._currentRow = len(self.df)
+            self.valueDicts = [nt._asdict() for nt in self.df.itertuples()]
+        else:
+            if os.path.exists(csvPath):
+                log.info(f"Results will be written to new file {csvPath}")
+            else:
+                log.warning(f"Results in existing file ({csvPath}) will be overwritten (non-incremental mode)")
+            self.df = None
+            self._currentRow = 0
+            self.valueDicts = []
 
     def addValues(self, values: Dict[str, Any]):
         """
@@ -134,20 +147,25 @@ class ParametersMetricsCollection:
         :return:
         """
         if self.df is None:
-            self.cols = list(values.keys())
+            cols = list(values.keys())
 
             # check sort column and move it to the front
             if self.sortColumnName is not None:
-                if self.sortColumnName not in self.cols:
-                    log.warning(f"Specified sort column '{self.sortColumnName}' not in list of columns: {self.cols}; sorting will not take place!")
+                if self.sortColumnName not in cols:
+                    log.warning(f"Specified sort column '{self.sortColumnName}' not in list of columns: {cols}; sorting will not take place!")
                 else:
-                    self.cols.remove(self.sortColumnName)
-                    self.cols.insert(0, self.sortColumnName)
+                    cols.remove(self.sortColumnName)
+                    cols.insert(0, self.sortColumnName)
 
-            self.df = pd.DataFrame(columns=self.cols)
+            self.df = pd.DataFrame(columns=cols)
+        else:
+            # check for new columns
+            for col in values.keys():
+                if col not in self.df.columns:
+                    self.df[col] = None
 
         # append data to data frame
-        self.df.loc[self._currentRow] = [values[c] for c in self.cols]
+        self.df.loc[self._currentRow] = [values.get(c) for c in self.df.columns]
         self._currentRow += 1
 
         # sort where applicable
@@ -167,8 +185,19 @@ class ParametersMetricsCollection:
     def getDataFrame(self) -> pd.DataFrame:
         return self.df
 
+    def contains(self, values: Dict[str, Any]):
+        for existingValues in self.valueDicts:
+            isContained = True
+            for k, v in values.items():
+                ev = existingValues.get(k)
+                if ev != v and str(ev) != str(v):
+                    isContained = False
+                    break
+            if isContained:
+                return True
 
-class GridSearch(TrackedExperimentDataProvider):
+
+class GridSearch(TrackingMixin):
     """
     Instances of this class can be used for evaluating models with different user-provided parametrizations
     over the same data and persisting the results
@@ -176,15 +205,26 @@ class GridSearch(TrackedExperimentDataProvider):
     log = log.getChild(__qualname__)
 
     def __init__(self, modelFactory: Callable[..., VectorModel], parameterOptions: Union[Dict[str, Sequence[Any]], List[Dict[str, Sequence[Any]]]],
-            numProcesses=1, csvResultsPath: str = None, parameterCombinationSkipDecider: ParameterCombinationSkipDecider = None):
+            numProcesses=1, csvResultsPath: str = None, incremental=False, incrementalSkipExisting=False,
+            parameterCombinationSkipDecider: ParameterCombinationSkipDecider = None, modelSaveDirectory: str = None,
+            name: str = None):
         """
         :param modelFactory: the function to call with keyword arguments reflecting the parameters to try in order to obtain a model instance
         :param parameterOptions: a dictionary which maps from parameter names to lists of possible values - or a list of such dictionaries,
             where each dictionary in the list has the same keys
         :param numProcesses: the number of parallel processes to use for the search (use 1 to run without multi-processing)
-        :param csvResultsPath: the path of a CSV file to which the results shall be written
+        :param csvResultsPath: the path to a directory or concrete CSV file to which the results shall be written;
+            if it is None, no CSV data will be written; if it is a directory, a file name starting with this grid search's name (see below)
+            will be created.
+            The resulting CSV data will contain one line per evaluated parameter combination.
+        :param incremental: whether to add to an existing CSV file instead of overwriting it
+        :param incrementalSkipExisting: if incremental mode is on, whether to skip any parameter combinations that are already present
+            in the CSV file
         :param parameterCombinationSkipDecider: an instance to which parameters combinations can be passed in order to decide whether the
             combination shall be skipped (e.g. because it is redundant/equivalent to another combination or inadmissible)
+        :param modelSaveDirectory: the directory where the serialized models shall be saved; if None, models are not saved
+        :param name: the name of this grid search, which will, in particular, be prepended to all saved model files;
+            if None, a default name will be generated of the form "gridSearch_<timestamp>"
         """
         self.modelFactory = modelFactory
         if type(parameterOptions) == list:
@@ -196,8 +236,14 @@ class GridSearch(TrackedExperimentDataProvider):
         else:
             self.parameterOptionsList = [parameterOptions]
         self.numProcesses = numProcesses
-        self.csvResultsPath = csvResultsPath
         self.parameterCombinationSkipDecider = parameterCombinationSkipDecider
+        self.modelSaveDirectory = modelSaveDirectory
+        self.name = name if name is not None else "gridSearch_" + datetime.now().strftime('%Y%m%d-%H%M%S')
+        self.csvResultsPath = csvResultsPath
+        self.incremental = incremental
+        self.incrementalSkipExisting = incrementalSkipExisting
+        if self.csvResultsPath is not None and os.path.isdir(csvResultsPath):
+            self.csvResultsPath = os.path.join(self.csvResultsPath, f"{self.name}_results.csv")
 
         self.numCombinations = 0
         for parameterOptions in self.parameterOptionsList:
@@ -208,10 +254,11 @@ class GridSearch(TrackedExperimentDataProvider):
         log.info(f"Created GridSearch object for {self.numCombinations} parameter combinations")
 
         self._executor = None
-        self._trackedExperiment = None
 
     @classmethod
-    def _evalParams(cls, modelFactory, metricsEvaluator: MetricsDictProvider, skipDecider: ParameterCombinationSkipDecider, **params) -> Optional[Dict[str, Any]]:
+    def _evalParams(cls, modelFactory: Callable[..., VectorModel], metricsEvaluator: MetricsDictProvider,
+            skipDecider: ParameterCombinationSkipDecider, gridSearchName, combinationIdx,
+            modelSaveDirectory: Optional[str], **params) -> Optional[Dict[str, Any]]:
         if skipDecider is not None:
             if skipDecider.isSkipped(params):
                 cls.log.info(f"Parameter combination is skipped according to {skipDecider}: {params}")
@@ -220,15 +267,17 @@ class GridSearch(TrackedExperimentDataProvider):
         model = modelFactory(**params)
         # TODO or not TODO: for some evaluators additional kwargs can be passed, e.g. onTrainingData
         values = metricsEvaluator.computeMetrics(model)
+        if modelSaveDirectory is not None:
+            filename = f"{gridSearchName}_{model.__class__.__name__}_{combinationIdx}_{uuid.uuid4()}.pickle"
+            log.info(f"Saving trained model to {filename} ...")
+            model.save(os.path.join(modelSaveDirectory, filename))
+            values["filename"] = filename
         values["str(model)"] = str(model)
         values.update(**params)
         if skipDecider is not None:
             skipDecider.tell(params, values)
         return values
 
-    def setTrackedExperiment(self, trackedExperiment: TrackedExperiment):
-        self._trackedExperiment = trackedExperiment
-
     def run(self, metricsEvaluator: MetricsDictProvider, sortColumnName=None, ascending=True) -> pd.DataFrame:
         """
         Run the grid search. If csvResultsPath was provided in the constructor, each evaluation result will be saved
@@ -240,9 +289,14 @@ class GridSearch(TrackedExperimentDataProvider):
         :param ascending: whether to sort in ascending order; has an effect only if sortColumnName is not None
         :return: the data frame with all evaluation results
         """
-        loggingCallback = self._trackedExperiment.trackValues if self._trackedExperiment is not None else None
-        paramsMetricsCollection = ParametersMetricsCollection(csvPath=self.csvResultsPath,
-                                                  sortColumnName=sortColumnName, ascending=ascending)
+        if self.trackedExperiment is not None:
+            loggingCallback = self.trackedExperiment.trackValues
+        elif metricsEvaluator.trackedExperiment is not None:
+            loggingCallback = metricsEvaluator.trackedExperiment.trackValues
+        else:
+            loggingCallback = None
+        paramsMetricsCollection = ParametersMetricsCollection(csvPath=self.csvResultsPath, sortColumnName=sortColumnName,
+            ascending=ascending, incremental=self.incremental)
 
         def collectResult(values):
             if values is None:
@@ -253,23 +307,36 @@ class GridSearch(TrackedExperimentDataProvider):
             log.info(f"Updated grid search result:\n{paramsMetricsCollection.getDataFrame().to_string()}")
 
         if self.numProcesses == 1:
+            combinationIdx = 0
             for parameterOptions in self.parameterOptionsList:
                 for paramsDict in iterParamCombinations(parameterOptions):
-                    collectResult(self._evalParams(self.modelFactory, metricsEvaluator, self.parameterCombinationSkipDecider, **paramsDict))
+                    if self.incrementalSkipExisting and self.incremental:
+                        if paramsMetricsCollection.contains(paramsDict):
+                            log.info(f"Skipped because parameters are already present in collection (incremental mode): {paramsDict}")
+                            continue
+                    collectResult(self._evalParams(self.modelFactory, metricsEvaluator, self.parameterCombinationSkipDecider, self.name,
+                        combinationIdx, self.modelSaveDirectory, **paramsDict))
+                    combinationIdx += 1
         else:
             executor = ProcessPoolExecutor(max_workers=self.numProcesses)
             futures = []
+            combinationIdx = 0
             for parameterOptions in self.parameterOptionsList:
                 for paramsDict in iterParamCombinations(parameterOptions):
+                    if self.incrementalSkipExisting and self.incremental:
+                        if paramsMetricsCollection.contains(paramsDict):
+                            log.info(f"Skipped because parameters are already present in collection (incremental mode): {paramsDict}")
+                            continue
                     futures.append(executor.submit(self._evalParams, self.modelFactory, metricsEvaluator, self.parameterCombinationSkipDecider,
-                                                   **paramsDict))
+                        self.name, combinationIdx, self.modelSaveDirectory, **paramsDict))
+                    combinationIdx += 1
             for future in futures:
                 collectResult(future.result())
 
         return paramsMetricsCollection.getDataFrame()
 
 
-class SAHyperOpt(TrackedExperimentDataProvider):
+class SAHyperOpt(TrackingMixin):
     log = log.getChild(__qualname__)
 
     class State(SAState):
@@ -311,12 +378,12 @@ class SAHyperOpt(TrackedExperimentDataProvider):
             pass
 
     def __init__(self, modelFactory: Callable[..., VectorModel],
-                 opsAndWeights: List[Tuple[Callable[['SAHyperOpt.State'], 'SAHyperOpt.ParameterChangeOperator'], float]],
-                 initialParameters: Dict[str, Any], metricsEvaluator: MetricsDictProvider,
-                 metricToOptimise, minimiseMetric=False,
-                 collectDataFrame=True, csvResultsPath: Optional[str] = None,
-                 parameterCombinationEquivalenceClassValueCache: ParameterCombinationEquivalenceClassValueCache = None,
-                 p0=0.5, p1=0.0):
+            opsAndWeights: List[Tuple[Callable[['SAHyperOpt.State'], 'SAHyperOpt.ParameterChangeOperator'], float]],
+            initialParameters: Dict[str, Any], metricsEvaluator: MetricsDictProvider,
+            metricToOptimise, minimiseMetric=False,
+            collectDataFrame=True, csvResultsPath: Optional[str] = None,
+            parameterCombinationEquivalenceClassValueCache: ParameterCombinationEquivalenceClassValueCache = None,
+            p0=0.5, p1=0.0):
         """
         :param modelFactory: a factory for the generation of models which is called with the current parameter combination
             (all keyword arguments), initially initialParameters
@@ -350,20 +417,19 @@ class SAHyperOpt(TrackedExperimentDataProvider):
         self.p0 = p0
         self.p1 = p1
         self._sa = None
-        self._trackedExperiment = None
-
-    def setTrackedExperiment(self, trackedExperiment: TrackedExperiment):
-        self._trackedExperiment = trackedExperiment
 
     @classmethod
     def _evalParams(cls, modelFactory, metricsEvaluator: MetricsDictProvider, parametersMetricsCollection: Optional[ParametersMetricsCollection],
-                    parameterCombinationEquivalenceClassValueCache, trackedExperiment, **params):
+            parameterCombinationEquivalenceClassValueCache, trackedExperiment, **params):
+        if trackedExperiment is not None and metricsEvaluator.trackedExperiment is not None:
+            log.warning(f"Tracked experiment already set in evaluator, results will be tracked twice and"
+                        f"might get overwritten!")
+
         metrics = None
         if parameterCombinationEquivalenceClassValueCache is not None:
             metrics = parameterCombinationEquivalenceClassValueCache.get(params)
         if metrics is not None:
             cls.log.info(f"Result for parameter combination {params} could be retrieved from cache, not adding new result")
-            return metrics
         else:
             cls.log.info(f"Evaluating parameter combination {params}")
             model = modelFactory(**params)
@@ -384,7 +450,7 @@ class SAHyperOpt(TrackedExperimentDataProvider):
 
     def _computeMetric(self, params):
         metrics = self._evalParams(self.modelFactory, self.evaluatorOrValidator, self.parametersMetricsCollection,
-            self.parameterCombinationEquivalenceClassValueCache, self._trackedExperiment, **params)
+            self.parameterCombinationEquivalenceClassValueCache, self.trackedExperiment, **params)
         metricValue = metrics[self.metricToOptimise]
         if not self.minimiseMetric:
             return -metricValue
